from django.test import TestCase, RequestFactory
from django.urls import reverse

from ..models import AccessToken, AccessTokenUse
from ..views import resolve_return_url, gateway


class PerimeterViewTests(TestCase):
    def setUp(self):
        self.factory = RequestFactory()
        self.url = reverse("perimeter:gateway")

    def test_display_gateway_GET_200(self):
        """GET on the perimeter:gateway should work."""
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "perimeter/gateway.html")

    def test_display_gateway_POST_302(self):
        """POST on the perimeter:gateway should work."""
        token = AccessToken.objects.create_access_token()
        payload = {"token": token.token}
        request = self.factory.post(self.url, payload)
        request.session = {}
        response = gateway(request)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response["location"], self.url)
        usage = AccessTokenUse.objects.get()
        self.assertIsNone(usage.user_email)
        self.assertIsNone(usage.user_name)
        self.assertEqual(usage.token, token)

        # Check the next url is decoded and used properly
        url = self.url + "?next=%2Fadmin%2F%3Fimportant%3Dparam"
        request = self.factory.post(url, payload)
        request.session = {}
        response = gateway(request)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response["location"], "/admin/?important=param")

    def test_resolve_return_url(self):
        default_url = reverse("perimeter:gateway")
        for url in (None, "x/y/z/", default_url):
            self.assertEqual(resolve_return_url(url), default_url)



"""
Middleware component of Perimeter app - checks all incoming requests for a
valid token. See Perimeter docs for more details.
"""
from urllib.parse import urlencode

from django.core.exceptions import ImproperlyConfigured, MiddlewareNotUsed
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.deprecation import MiddlewareMixin

from .models import AccessToken
from .settings import (
    PERIMETER_BYPASS_FUNCTION as bypass_perimeter,
    PERIMETER_ENABLED,
    HTTP_X_PERIMETER_TOKEN,
    PERIMETER_SESSION_KEY,
)


def check_middleware(func):
    """Simple decorator that checks a request arg has a Session attached."""

    def inner(request, *args):
        if not hasattr(request, "session"):
            raise ImproperlyConfigured(
                "Missing session attribute - please check MIDDLEWARE_CLASSES for "
                "'django.contrib.sessions.middleware.SessionMiddleware'."
            )
        return func(request, *args)

    return inner


@check_middleware
def get_request_token(request):
    """Extract token string from HTTP header or querystring."""
    return request.META.get(HTTP_X_PERIMETER_TOKEN, None) or request.session.get(
        PERIMETER_SESSION_KEY, None
    )


@check_middleware
def set_request_token(request, token_value):
    """Sets the request.session token value.

    Args:
        token - string, the token value (not the token object, as that is
            not serializable)
    """
    request.session[PERIMETER_SESSION_KEY] = token_value


def get_access_token(request):
    """Fetch the AccessToken from the request."""
    token_value = get_request_token(request)
    return AccessToken.objects.get_access_token(token_value)


def get_redirect_url(request):
    """Unpack request and extract a valid redirect url."""
    qstring = urlencode({"next": request.get_full_path()})
    url = reverse("perimeter:gateway")
    return f"{url}?{qstring}"


class PerimeterAccessMiddleware(MiddlewareMixin):
    """
    Middleware used to detect whether user can access site or not.

    This middleware will be disabled if the PERIMETER_ENABLED setting does not
    exist in django settings, or is False.
    """

    def __init__(self, *args, **kwargs):
        """
        Disable middleware if PERIMETER_ENABLED setting not True.

        Raises MiddlewareNotUsed exception if the PERIMETER_ENABLED setting
        is not True - this is used by Django framework to remove the middleware.
        """
        if PERIMETER_ENABLED is False:
            raise MiddlewareNotUsed("Perimeter disabled")
        super().__init__(*args, **kwargs)

    def process_request(self, request):
        """Check user session for token."""
        if bypass_perimeter(request):
            return None

        if get_access_token(request).is_valid:
            return None

        return HttpResponseRedirect(get_redirect_url(request))



# -*- coding: utf-8 -*-


from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [("perimeter", "0001_initial")]

    operations = [
        migrations.AlterField(
            model_name="accesstoken",
            name="token",
            field=models.CharField(unique=True, max_length=50),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name="accesstokenuse",
            name="user_email",
            field=models.EmailField(
                max_length=75, verbose_name="Token used by (email)"
            ),
            preserve_default=True,
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-04-07 06:39


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("perimeter", "0003_auto_20150612_0151")]

    operations = [
        migrations.AlterField(
            model_name="accesstokenuse",
            name="user_email",
            field=models.EmailField(
                blank=True,
                max_length=254,
                null=True,
                verbose_name="Token used by (email)",
            ),
        ),
        migrations.AlterField(
            model_name="accesstokenuse",
            name="user_name",
            field=models.CharField(
                blank=True,
                max_length=100,
                null=True,
                verbose_name="Token used by (name)",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2018-05-20 10:37
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("perimeter", "0004_auto_20170407_0639")]

    operations = [
        migrations.AlterField(
            model_name="accesstokenuse",
            name="client_user_agent",
            field=models.TextField(blank=True, verbose_name="Client User Agent"),
        )
    ]



from datetime import datetime, date, time, timedelta

from django.core.cache import cache
from django.test import TestCase
from django.utils.timezone import (
    now,
    is_aware,
    is_naive,
    make_aware,
    get_current_timezone,
)

from ..models import AccessToken, AccessTokenUse, default_expiry, EmptyToken
from ..settings import PERIMETER_DEFAULT_EXPIRY

TODAY = now().date()
YESTERDAY = TODAY - timedelta(days=1)
TOMORROW = TODAY + timedelta(days=1)


class EmptyTokenTests(TestCase):
    def test_is_valid(self):
        """EmptyToken should always be invalid."""
        token = EmptyToken()
        self.assertFalse(token.is_valid)


class AccessTokenManagerTests(TestCase):
    def test_create_with_token(self):
        """If a token is passed in to create_access_token, it's used."""
        token = AccessToken.objects.create_access_token(token="x")
        self.assertEqual(token, AccessToken.objects.get())
        self.assertTrue(token, "x")

    def test_create_with_expires(self):
        """If an expires is passed in to create_access_token, it's used."""
        tomorrow = date.today() + timedelta(days=1)
        token = AccessToken.objects.create_access_token(expires_on=tomorrow)
        self.assertEqual(token, AccessToken.objects.get())
        self.assertTrue(token.expires_on, tomorrow)

    def test_create_without_token(self):
        """If no token is passed in to create_access_token, a random one is used."""
        token = AccessToken.objects.create_access_token()
        self.assertEqual(token, AccessToken.objects.get())
        self.assertTrue(len(token.token), 10)

    def test_get_access_token(self):
        """Test the caching works."""
        token = AccessToken.objects.create_access_token()
        cache.clear()
        self.assertIsNone(cache.get(token.cache_key))
        token2 = AccessToken.objects.get_access_token(token.token)
        self.assertEqual(token, token2)
        self.assertIsNotNone(cache.get(token.cache_key))


class AccessTokenTests(TestCase):
    def test_default_expiry(self):
        self.assertEqual(
            default_expiry(), now().date() + timedelta(days=PERIMETER_DEFAULT_EXPIRY)
        )

    def test_attrs(self):
        # start with the defaults
        at = AccessToken()
        self.assertEqual(at.token, "")
        self.assertEqual(at.is_active, True)
        self.assertEqual(at.expires_on, default_expiry())
        self.assertEqual(at.created_at, None)
        self.assertEqual(at.updated_at, None)
        self.assertEqual(at.created_by, None)

        # check the timestamps
        at = at.save()
        self.assertIsNotNone(at.created_at)
        self.assertEqual(at.updated_at, at.created_at)

        # check the timestamps are updated
        x = at.created_at
        at = at.save()
        # created_at is _not_ updated
        self.assertEqual(at.created_at, x)
        # but updated_at _is_
        self.assertTrue(at.updated_at > at.created_at)

        self.assertTrue(at.is_valid)
        self.assertFalse(at.has_expired)

    def test_str(self):
        today = date.today()
        at = AccessToken(token="¡€#¢∞§¶•ªº", expires_on=today, is_active=True)
        self.assertEqual(str(at), "¡€#¢∞§¶•ªº")

    def test_cache_key(self):
        token = AccessToken(token="test")
        self.assertIsNotNone(token.cache_key)
        self.assertEqual(token.cache_key, AccessToken.get_cache_key("test"))

    def test_cache_management(self):
        token = AccessToken.objects.create_access_token()
        self.assertEqual(cache.get(token.cache_key), token)
        token.delete()
        self.assertIsNone(cache.get(token.cache_key))

    def test_generate_random_token(self):
        f = AccessToken._meta.get_field("token").max_length
        t1 = AccessToken.random_token_value()
        t2 = AccessToken.random_token_value()
        self.assertNotEqual(t1, t2)
        self.assertEqual(len(t1), f)

    def test_has_expired(self):
        at = AccessToken()
        at.expires_on = YESTERDAY
        self.assertTrue(at.has_expired)
        at.expires_on = TODAY
        self.assertFalse(at.has_expired)
        at.expires_on = TOMORROW
        self.assertFalse(at.has_expired)

    def test_seconds_to_expiry(self):
        "Test that it handles naive and tz-aware times"

        with self.settings(USE_TZ=False):
            at = AccessToken(expires_on=TOMORROW)
            expires_at = datetime.combine(at.expires_on, time.min)
            self.assertTrue(is_naive(expires_at))
            self.assertTrue(is_naive(now()))
            self.assertEqual(
                at.seconds_to_expiry, int((expires_at - now()).total_seconds())
            )

        with self.settings(USE_TZ=True):
            at = AccessToken(expires_on=TOMORROW)
            expires_at = make_aware(
                datetime.combine(at.expires_on, time.min), get_current_timezone()
            )
            self.assertTrue(is_aware(expires_at))
            self.assertTrue(is_aware(now()))
            self.assertEqual(
                at.seconds_to_expiry, int((expires_at - now()).total_seconds())
            )

    def test_is_valid(self):
        def assertValidity(active, expires, valid):
            return AccessToken(is_active=True, expires_on=TOMORROW).is_valid

        assertValidity(True, YESTERDAY, False)
        assertValidity(True, TODAY, True)
        assertValidity(True, TOMORROW, True)

        assertValidity(False, YESTERDAY, False)
        assertValidity(False, TODAY, False)
        assertValidity(False, TOMORROW, False)

    def test_record(self):
        at = AccessToken(token="test_token").save()
        atu = at.record("hugo@yunojuno.com", "Hugo")
        self.assertEqual(atu, AccessTokenUse.objects.get())
        self.assertEqual(atu.user_email, "hugo@yunojuno.com")
        self.assertEqual(atu.user_name, "Hugo")
        self.assertIsNotNone(atu.timestamp, "Hugo")
        self.assertEqual(atu.client_ip, "unknown")
        self.assertEqual(atu.client_user_agent, "unknown")


class AccesTokenUseTests(TestCase):
    def setUp(self):
        self.token = AccessToken(token="foo").save()

    def test_save(self):
        atu = AccessTokenUse(token=self.token)
        atu.save()
        self.assertIsNone(atu.user_email)
        self.assertIsNone(atu.user_name)



from os import environ

from django.test import TestCase

from ..settings import get_setting, CAST_AS_INT, CAST_AS_BOOL


class SettingsTests(TestCase):
    def test_cast_as_bool(self):

        self.assertTrue(CAST_AS_BOOL(True))
        self.assertTrue(CAST_AS_BOOL("True"))
        self.assertTrue(CAST_AS_BOOL("true"))
        # special case: 1 == True
        self.assertTrue(CAST_AS_BOOL(1))
        self.assertFalse(CAST_AS_BOOL(False))
        self.assertFalse(CAST_AS_BOOL(None))
        self.assertFalse(CAST_AS_BOOL("1"))

    def test_get_settings(self):

        with self.settings(TEST_SETTING=True):
            self.assertTrue(get_setting("TEST_SETTING", False))
            self.assertTrue(get_setting("TEST_SETTING", True))

        with self.settings(TEST_SETTING="0"):
            self.assertEqual(get_setting("TEST_SETTING", False), "0")

        with self.settings(TEST_SETTING="0"):
            self.assertEqual(
                get_setting("TEST_SETTING", False, cast_func=CAST_AS_INT), 0
            )

        with self.settings(TEST_SETTING="True"):
            self.assertEqual(
                get_setting("TEST_SETTING", False, cast_func=CAST_AS_BOOL), True
            )

        # confirm that env vars trump django.conf
        environ["TEST_SETTING"] = "True"
        with self.settings(TEST_SETTING=False):
            self.assertEqual(
                get_setting("TEST_SETTING", False, cast_func=CAST_AS_BOOL), True
            )



# -*- coding: utf-8 -*-


from django.db import models, migrations
from django.conf import settings
import perimeter.models


class Migration(migrations.Migration):

    dependencies = [migrations.swappable_dependency(settings.AUTH_USER_MODEL)]

    operations = [
        migrations.CreateModel(
            name="AccessToken",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("token", models.CharField(unique=True, max_length=10)),
                ("is_active", models.BooleanField(default=True)),
                (
                    "expires_on",
                    models.DateField(default=perimeter.models.default_expiry),
                ),
                ("created_at", models.DateTimeField()),
                ("updated_at", models.DateTimeField()),
                (
                    "created_by",
                    models.ForeignKey(
                        to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE, null=True
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="AccessTokenUse",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "user_email",
                    models.EmailField(
                        max_length=254, verbose_name="Token used by (email)"
                    ),
                ),
                (
                    "user_name",
                    models.CharField(
                        max_length=100, verbose_name="Token used by (name)"
                    ),
                ),
                (
                    "client_ip",
                    models.CharField(
                        max_length=15, verbose_name="IP address", blank=True
                    ),
                ),
                (
                    "client_user_agent",
                    models.CharField(
                        max_length=150, verbose_name="User Agent", blank=True
                    ),
                ),
                ("timestamp", models.DateTimeField()),
                (
                    "token",
                    models.ForeignKey(
                        to="perimeter.AccessToken", on_delete=models.CASCADE
                    ),
                ),
            ],
        ),
    ]
